@@ -24,12 +24,16 @@
 #include <ksharedconfig.h>
 
 #include <Kirigami/PlatformTheme>
+#include <Kirigami/TabletModeWatcher>
 
 QStyle *KQuickStyleItem::s_style = nullptr;
 
 QStyle *KQuickStyleItem::style()
 {
-    auto style = qApp->style();
+    if (!qobject_cast<QApplication *>(QCoreApplication::instance())) {
+        return s_style;
+    }
+    QStyle *style = qApp->style();
     return style ? style : s_style;
 }
 
@@ -59,10 +63,37 @@ KQuickStyleItem::KQuickStyleItem(QQuickItem *parent)
     , m_textureHeight(0)
     , m_lastFocusReason(Qt::NoFocusReason)
 {
-    // There is no styleChanged signal and QApplication sends QEvent::StyleChange only to all QWidgets
-    if (qApp->style()) {
-        connect(qApp->style(), &QObject::destroyed, this, &KQuickStyleItem::styleChanged);
+    // Check that we really are a QApplication before attempting to access the
+    // application style.
+    //
+    // Functions used in this file which are safe to access through qApp are:
+    //
+    //   qApp->font() and qApp->fontChanged() - provided by QGuiApplication
+    //   qApp->font("classname") - provided by QApplication but safe
+    //   qApp->layoutDirection() - provided by QGuiApplication
+    //   qApp->wheelScrollLines() - uses styleHints() provided by QGuiApplication
+    //   qApp->testAttribute() and qApp->setAttribute() - provided by QCoreApplication
+    //   qApp->devicePixelRatio() - provided by QGuiApplication
+    //   qApp->palette("classname") - provided by QApplication but safe
+    //
+    // but any use of qApp->style(), even if only trying to test that it exists,
+    // will assert.  Use KQuickStyleItem::style() as above.
+    //
+    // Any use of any other function provided by QApplication must either be checked
+    // to ensure that it is safe to use if not a QApplication, or guarded.
+
+    if (qobject_cast<QApplication *>(QCoreApplication::instance())) {
+        // We are a QApplication.  Unfortunately there is no styleChanged signal
+        // available, and it only sends QEvent::StyleChange events to all QWidgets.
+        // So watch for the existing application style being destroyed, which means
+        // that a new one is being applied.  See QApplication::setStyle().
+        QStyle *style = qApp->style();
+        if (style) {
+            connect(style, &QObject::destroyed, this, &KQuickStyleItem::styleChanged);
+        }
     } else {
+        // We are not a QApplication.  Create an internal copy of the configured
+        // desktop style, to be used for metrics, options and painting.
         KSharedConfig::Ptr kdeglobals = KSharedConfig::openConfig();
         KConfigGroup cg(kdeglobals, "KDE");
         auto style = s_style;
@@ -76,41 +107,9 @@ KQuickStyleItem::KQuickStyleItem(QQuickItem *parent)
     setFlag(QQuickItem::ItemHasContents, true);
     setSmooth(false);
 
-    connect(this, &KQuickStyleItem::visibleChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::widthChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::heightChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::enabledChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::infoChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::onChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::selectedChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::activeChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::textChanged, this, &KQuickStyleItem::updateSizeHint);
-    connect(this, &KQuickStyleItem::textChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::activeChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::flatChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::sunkenChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::hoverChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::maximumChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::minimumChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::valueChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::horizontalChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::transientChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::activeControlChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::hasFocusChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::activeControlChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::hintChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::propertiesChanged, this, &KQuickStyleItem::updateSizeHint);
-    connect(this, &KQuickStyleItem::propertiesChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::elementTypeChanged, this, &KQuickStyleItem::updateItem);
-    connect(this, &KQuickStyleItem::contentWidthChanged, this, &KQuickStyleItem::updateSizeHint);
-    connect(this, &KQuickStyleItem::contentHeightChanged, this, &KQuickStyleItem::updateSizeHint);
-    connect(this, &KQuickStyleItem::widthChanged, this, &KQuickStyleItem::updateRect);
-    connect(this, &KQuickStyleItem::heightChanged, this, &KQuickStyleItem::updateRect);
-
-    connect(this, &KQuickStyleItem::heightChanged, this, &KQuickStyleItem::updateBaselineOffset);
-    connect(this, &KQuickStyleItem::contentHeightChanged, this, &KQuickStyleItem::updateBaselineOffset);
-
     connect(qApp, &QApplication::fontChanged, this, &KQuickStyleItem::updateSizeHint, Qt::QueuedConnection);
+
+    Kirigami::TabletModeWatcher::self()->addWatcher(this);
 }
 
 KQuickStyleItem::~KQuickStyleItem()
@@ -150,6 +149,7 @@ KQuickStyleItem::~KQuickStyleItem()
     }
 
     m_styleoption = nullptr;
+    Kirigami::TabletModeWatcher::self()->removeWatcher(this);
 }
 
 void KQuickStyleItem::initStyleOption()
@@ -721,7 +721,8 @@ void KQuickStyleItem::initStyleOption()
     }
 
     m_styleoption->styleObject = this;
-    m_styleoption->direction = qApp->layoutDirection();
+    const auto mirror = m_control == nullptr ? qApp->layoutDirection() == Qt::RightToLeft : m_control->property("mirrored").toBool();
+    m_styleoption->direction = mirror ? Qt::RightToLeft : Qt::LeftToRight;
 
     int w = m_textureWidth > 0 ? m_textureWidth : width();
     int h = m_textureHeight > 0 ? m_textureHeight : height();
@@ -776,6 +777,7 @@ void KQuickStyleItem::initStyleOption()
     } else if (sizeHint == QLatin1String("small")) {
         m_styleoption->state |= QStyle::State_Small;
     }
+
 }
 
 QIcon KQuickStyleItem::iconFromIconProperty() const
@@ -1233,6 +1235,7 @@ void KQuickStyleItem::setContentWidth(int arg)
 {
     if (m_contentWidth != arg) {
         m_contentWidth = arg;
+        updateSizeHint();
         Q_EMIT contentWidthChanged(arg);
     }
 }
@@ -1241,6 +1244,8 @@ void KQuickStyleItem::setContentHeight(int arg)
 {
     if (m_contentHeight != arg) {
         m_contentHeight = arg;
+        updateSizeHint();
+        updateBaselineOffset();
         Q_EMIT contentHeightChanged(arg);
     }
 }
@@ -1365,6 +1370,7 @@ void KQuickStyleItem::setHints(const QVariantMap &str)
         m_hints = str;
         initStyleOption();
         updateSizeHint();
+        polish();
         if (m_styleoption->state & QStyle::State_Mini) {
             m_font.setPointSize(9.);
             Q_EMIT fontChanged();
@@ -1472,6 +1478,7 @@ void KQuickStyleItem::setElementType(const QString &str)
     Q_EMIT topPaddingChanged();
     Q_EMIT bottomPaddingChanged();
     updateSizeHint();
+    polish();
 }
 
 QRectF KQuickStyleItem::subControlRect(const QString &subcontrolString)
@@ -1780,7 +1787,16 @@ bool KQuickStyleItem::event(QEvent *ev)
             polish();
         }
         return true;
+    } else if (ev->type() == Kirigami::TabletModeChangedEvent::type) {
+        Q_EMIT leftPaddingChanged();
+        Q_EMIT rightPaddingChanged();
+        Q_EMIT topPaddingChanged();
+        Q_EMIT bottomPaddingChanged();
+        updateSizeHint();
+        polish();
+        return true;
     }
+
     return QQuickItem::event(ev);
 }
 
@@ -1918,15 +1934,48 @@ bool KQuickStyleItem::eventFilter(QObject *watched, QEvent *event)
     return QQuickItem::eventFilter(watched, event);
 }
 
+void KQuickStyleItem::itemChange(QQuickItem::ItemChange change, const QQuickItem::ItemChangeData &value)
+{
+    if (change == QQuickItem::ItemVisibleHasChanged || change == QQuickItem::ItemEnabledHasChanged) {
+        polish();
+    }
+
+    QQuickItem::itemChange(change, value);
+}
+
+#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
+void KQuickStyleItem::geometryChanged(const QRectF &newGeometry, const QRectF &oldGeometry)
+{
+    QQuickItem::geometryChanged(newGeometry, oldGeometry);
+#else
+void KQuickStyleItem::geometryChange(const QRectF &newGeometry, const QRectF &oldGeometry)
+{
+    QQuickItem::geometryChange(newGeometry, oldGeometry);
+#endif
+
+    if (!newGeometry.isEmpty() && newGeometry != oldGeometry) {
+        polish();
+        updateRect();
+
+        if (newGeometry.height() != oldGeometry.height()) {
+            updateBaselineOffset();
+        }
+    }
+}
+
 void KQuickStyleItem::styleChanged()
 {
-    if (!qApp->style() || QApplication::closingDown()) {
+    // It should be safe to use qApp->style() unguarded here, because the signal
+    // will only have been connected if qApp is a QApplication.
+    Q_ASSERT(qobject_cast<QApplication *>(QCoreApplication::instance()));
+    auto *style = qApp->style();
+    if (!style || QCoreApplication::closingDown()) {
         return;
     }
 
-    Q_ASSERT(qApp->style() != sender());
+    Q_ASSERT(style != sender());
 
-    connect(qApp->style(), &QObject::destroyed, this, &KQuickStyleItem::styleChanged);
+    connect(style, &QObject::destroyed, this, &KQuickStyleItem::styleChanged);
 
     updateSizeHint();
     updateItem();
